#ifndef _CONVERTER_H_
#define _CONVERTER_H_
/*----------------------------------------------------------------------*\

  CONVERTER.H

  Header file for converter unit of ARUN Alan System interpreter

\*----------------------------------------------------------------------*/

/* Data: */

/* Functions: */
extern char *ensureExternalEncoding(char input[]);
extern char *ensureInternalEncoding(char string[]);

#endif
