#ifndef TI994ATERP_H
#define TI994ATERP_H

#include "scottdefines.h"

void RunImplicitTI99Actions(void);
ExplicitResultType RunExplicitTI99Actions(int verb_num, int noun_num);

#endif /* TI994ATERP_H */
