//
//  apple2detect.h
//  Part of ScottFree, an interpreter for adventures in Scott Adams format
//
//  Created by Petter Sjölund on 2022-08-03.
//

#ifndef apple2detect_h
#define apple2detect_h

#include <stdio.h>

#include "scottdefines.h"

GameIDType DetectApple2(uint8_t **sf, size_t *extent);

#endif /* apple2detect_h */
