//
//  gameinfo.h
//  Part of Plus, an interpreter for Scott Adams Graphic Adventures Plus
//
//  Created by Petter Sjölund on 2022-01-09.
//

#ifndef gameinfo_h
#define gameinfo_h

#include "definitions.h"

extern struct GameInfo games[];
extern const char *sysdict[];
extern const char *sysdict_i_am[];

#endif /* gameinfo_h */
