//
//  titleimage.h
//  part of ScottFree, an interpreter for adventures in Scott Adams format
//
//  Created by Petter Sjölund on 2022-10-01.
//

#ifndef titleimage_h
#define titleimage_h

#include <stdio.h>

void DrawTitleImage(void);
void PrintTitleScreenBuffer(void);
void PrintTitleScreenGrid(void);

#endif /* titleimage_h */
