//
//  debug.h
//  part of ScottFree, an interpreter for adventures in Scott Adams format
//
//  Created by Petter Sjölund on 2022-01-19.
//

#ifndef debug_h
#define debug_h

#include <stdio.h>

#endif /* debug_h */
