/* alan.version.c - Created by venum 4.2beta1 */

#include "alan.version.h"

Product alan = {
  "Alan",
  "Adventure Language System",
  "Alan 3.0beta8",
  "Alan 3.0beta8 -- Adventure Language System (2021-08-17 08:33)",
  "2021-08-17",
  "08:33:17",
  "",
  "thoni64",
  "linux",
  {"3.0beta8", 3, 0, 8, 1629181997, "beta"}
};

char *alanId =
  "@(#)RELEASE ";
