//
//  loadtotpicture.h
//
//  Part of TaylorMade, an interpreter for Adventure Soft UK games
//
//  This code is for de-protecting the ZX Spectrum text-only version of Temple of Terror
//
//  Created by Petter Sjölund on 2022-04-19.
//

#ifndef loadtotpicture_h
#define loadtotpicture_h

#include <stdint.h>

uint8_t *LoadAlkatrazPicture(uint8_t *mem, uint8_t *file);

#endif /* loadtotpicture_h */
