//
//  gremlins.h
//  part of ScottFree, an interpreter for adventures in Scott Adams format
//
//  Created by Petter Sjölund on 2022-01-10.
//

#ifndef gremlins_h
#define gremlins_h

#include <stdio.h>

void UpdateGremlinsAnimations(void);
void GremlinsLook(void);
void GremlinsAction(void);
void LoadExtraGermanGremlinsData(void);
void LoadExtraGermanGremlinsC64Data(void);
void LoadExtraSpanishGremlinsData(void);
void LoadExtraSpanishGremlinsC64Data(void);

#endif /* gremlins_h */
