#include "options.h"

bool verboseOption = false;
bool ignoreErrorOption = false;
bool debugOption = false;
bool traceSectionOption = false;
bool tracePushOption = false;
bool traceStackOption = false;
bool traceSourceOption = false;
bool traceInstructionOption = false;
bool transcriptOption = false;
bool commandLogOption = false;
bool statusLineOption = true;
bool regressionTestOption = false;
bool nopagingOption = false;
int encodingOption = 0;         /* 0 = ISO, 1 = UTF-8 */
