//
//  stdetect.h
//  Part of Plus, an interpreter for Scott Adams Graphic Adventures Plus
//
//  Created by Petter Sjölund on 2022-08-02.
//

#ifndef stdetect_h
#define stdetect_h

#include <stdio.h>

int DetectST(uint8_t **sf, size_t *extent);

#endif /* stdetect_h */
