#define IDC_STATIC (-1)

#define IDR_MANIFEST            1
#define IDR_ARUN                25000
#define IDR_ABOUT               27000

#define IDD_ABOUT               26000
#define ID_MENU_RESTART         26001
#define ID_MENU_SAVE            26002
#define ID_MENU_RECORD          26003
#define ID_MENU_PLAYBACK        26004
#define ID_MENU_TRANSCRIPT      26005
#define ID_MENU_RESTORE         26006
#define ID_MENU_ABOUT           26007
