//
//  scottgameinfo.h
//  Part of ScottFree, an interpreter for adventures in Scott Adams format
//
//  Created by Petter Sjölund on 2022-01-09.
//

#ifndef scottgameinfo_h
#define scottgameinfo_h

#include "scottdefines.h"

extern struct GameInfo games[];
extern const char *sysdict[];
extern const char *sysdict_i_am[];
extern const char *sysdict_zx[];
extern const char *sysdict_TI994A[];

#endif /* scottgameinfo_h */
