//
//  atari8c64draw.h
//  Part of ScottFree, an interpreter for adventures in Scott Adams format
//
//  Created by Petter Sjölund on 2022-09-09.
//

#ifndef atari8c64draw_h
#define atari8c64draw_h

int DrawAtariC64Image(USImage *image);

#endif /* atari8c64draw_h */
