//
//  animations.h
//  Part of TaylorMade, an interpreter for Adventure Soft UK games
//
//  Created by Petter Sjölund on 2022-03-24.
//

#ifndef animations_h
#define animations_h

#include <stdio.h>
void UpdateRebelAnimations(void);
void UpdateKaylethAnimations(void);
void StartAnimations(void);

#endif /* animations_h */
